#!/usr/bin/python
#
# Copyright 2008 Google Inc. All Rights Reserved.
#
# This file is part of Kamailio, a free SIP server.
#
# SPDX-License-Identifier: GPL-2.0-or-later
#
# Kamailio is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version
#
# Kamailio is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA

"""Empty __init__ for dbtextdb module.

This file is empty, it only exists to enable importing dbtextdb
"""

__author__ = 'herman@google.com (Herman Sheremetyev)'



#!/usr/bin/python3

#parses the output of /usr/sbin/kamctl rpc app_python.api_list
#
#usage ./kemi_mock.py api.json > KSR.py
#or for python 3.2
#./kemi_mock.py api.json --no-union > KSR.py

import json
import sys

from collections import defaultdict

#python 3.2 doesnt support types.Union
noUnion = False

reserved_keywords = {"async"}


def printMocReturn(module_name, func, indent):
    param_names = []
    param_list = []
    param_signature = ""
    if func['params'] is not None and func['params'] != 'none':
        param_list = func['params'].split(", ")
    i = 0

    for _ in param_list:
        param_names.append("param"+str(i))
        i = i + 1

    param_signature = ", ".join(param_names)

    prefix = ""
    for i in range(indent):
        prefix = prefix+"\t"

    print(prefix + "if \""+func['name']+"\" not in _mock_data['"+module_name+"']:")
    printDefaultReturn(func, indent+1)

    print(prefix + "node = _mock_data['"+module_name+"']['"+func['name']+"']")

    print(prefix + "if isinstance(node, types.FunctionType):")
    print(prefix + "\treturn node("+param_signature+")")

    for param in param_names:
        print(prefix + "if not isinstance(node, dict):")
        print(prefix + "\treturn node")
        print(prefix + "if str(" + param + ") in node:")
        print(prefix + "\tnode = node[str("+param+")]")
        print(prefix + "else:")
        printDefaultReturn(func, indent+1)

    print(prefix + "return node")


def printDefaultReturn(func, indent):
    prefix = ""
    for i in range(indent):
        prefix = prefix+"\t"

    if func['ret'] == "bool":
        print(prefix + "return True")
    elif func['ret'] == "int":
        print(prefix + "return 1")
    elif func['ret'] == "str":
        print(prefix + "return \"\"")
    elif func['ret'] == "xval":
        print(prefix + "return None")
    else:
        print(prefix + "return")


def printFunction(module_name, func, indent):
    params = ""
    log_params = ""
    if module_name == "":
        log_params = "\"" + func['name'] + "\""
    else:
        log_params = "\"" + module_name + "." + func['name'] + "\""

    log_format_params = "%s"

    param_list = []
    if func['params'] is not None and func['params'] != "none":
        param_list = func['params'].split(", ")
        i = 0
        for _ in param_list:
            if params != "":
                 params = params + ", "
            params = params + "param" + str(i) + ": " + param_list[i]
            log_params = log_params + ", param" + str(i)
            log_format_params = log_format_params + ", %s"
            i = i+1
    if len(param_list) > 0:
        log_params = "(" + log_params + ")"
    prefix = ""
    for i in range(indent):
        prefix = prefix+"\t"
    if indent > 0:
        print(prefix + "@staticmethod")
    if func['ret'] == "bool":
        print(prefix + "def " + func['name'] +"("+params+") -> bool:")
    elif func['ret'] == "int":
        print(prefix + "def " + func['name'] +"("+params+") -> int:")
    elif func['ret'] == "str":
        print(prefix + "def " + func['name'] + "(" + params + ") -> int:")
    elif func['ret'] == "xval":
        if noUnion:
            print(prefix + "def " + func['name'] + "(" + params + "):")
        else:
            print(prefix + "def " + func['name'] +"("+params+") -> Union[int, str, None]:")
    else:
        print(prefix + "def " + func['name'] +"("+params+"):")

    generate_function_doc(module_name, func, prefix)

    print(prefix + "\tprint(\"Calling " + log_format_params + "\" % "+log_params+")")
    printMocReturn(module_name, func, indent+1)
    print("")


def generate_function_doc(module_name, func, prefix):
    if documentation is not None and module_name in documentation:
        function_parts = func['name'].split("_")
        for i in range(len(function_parts), 0, -1):
            function_prefix = "_".join(function_parts[:i])
            if function_prefix in documentation[module_name]["functions"]:
                print(prefix + "\t\"\"\"")
                documentation_lines = documentation[module_name]["functions"][function_prefix].split("\n")
                for line in documentation_lines:
                    print(prefix + "\t" + line.replace("\\", "\\\\\\\\"))
                print(prefix + "\t\"\"\"")
                break


classes = defaultdict(list)

if len(sys.argv) < 2:
    print("Please specify the json file to parse")
    sys.exit(-1)

documentation = None
if len(sys.argv) > 2:
    for i in range(2,len(sys.argv)):
        if sys.argv[i] == "--no-union":
            noUnion = True
        else:
            with open(sys.argv[i]) as f:
                documentation = json.load(f)

if not noUnion:
    print("from typing import Union")

print("import sys")
print("import types")
print("_mock_data = {}")
print("")

with open(sys.argv[1]) as f:
    data = json.load(f)

for method in data['result']['methods']:
    classes[method['func']['module']].append(method['func'])

if "pv" not in classes:
    classes['pv'].append({'params': 'str',
                          'ret': 'xval',
                          'name': 'get'}
                         )
    classes['pv'].append({'params': 'str',
                          'ret': 'xval',
                          'name': 'gete'}
                         )
    classes['pv'].append({'params': 'str, int',
                          'ret': 'xval',
                          'name': 'getvn'}
                         )
    classes['pv'].append({'params': 'str, str',
                          'ret': 'xval',
                          'name': 'getvs'}
                         )
    classes['pv'].append({'params': 'str',
                          'ret': 'xval',
                          'name': 'getw'}
                         )
    classes['pv'].append({'params': 'str, int',
                          'ret': 'none',
                          'name': 'seti'}
                         )
    classes['pv'].append({'params': 'str, str',
                          'ret': 'none',
                          'name': 'sets'}
                         )
    classes['pv'].append({'params': 'str',
                          'ret': 'none',
                          'name': 'unset'}
                         )
    classes['pv'].append({'params': 'str',
                          'ret': 'none',
                          'name': 'is_null'}
                         )

for module_name, module in classes.items():
    if module_name != "":
        print("")
        print("class " + module_name.capitalize() + ":")

        for func in module:
            printFunction(module_name, func, 1)

for func in classes['']:
    print("")
    printFunction('', func, 0)


def document_module(module_name):
    if documentation is not None and module_name in documentation:
        print("\"\"\"")
        documentation_lines = documentation[module_name]["overview"].split("\n")
        for line in documentation_lines:
            print("" + line)
        print("\"\"\"")


for module_name in classes.keys():
    if module_name != "":
        if module_name in reserved_keywords:
            print("setattr(sys.modules[__name__], '" + module_name + "', " + module_name.capitalize() + "())")
        else:
            print(module_name + " = "+module_name.capitalize()+"()")
        document_module(module_name)

print("")

for module_name in classes.keys():
    print("_mock_data['" + module_name + "'] = {}")




import sys
import xmlrpclib

# Usage:  python xmlrpc_test2.py command [params...]
#
# python script for sending an xmlrpc command to ser's xmlrpc module.
# This script uses python xmlrpclib directly and expects the remote side to
# immediately close the connection after answering (broken xmlrpclib
# behaviour).
# There are 2 way to make it work with ser xmlrpc module: define a
# better transport class (that's what the xmlrpc_test.py script is doing) or
# change ser xmlrpc route so that it will close the connection after each
# processes xmlrpc request (e.g. add a drop -1 at the end).
#
# See also: xmlrpc_test.py (better version, using a redefined transport class).
#
# History:
# --------
#  2009-07-13  initial version (andrei)
#

XMLRPC_SERVER = "127.0.0.1"
XMLRPC_PORT = 5060


if len(sys.argv) < 2:
    sys.exit("Usage: " + sys.argv[0] + " rpc_command [args...]")

client = xmlrpclib.ServerProxy(
    "http://{0}:{1}".format(XMLRPC_SERVER, XMLRPC_PORT))
res = getattr(client, sys.argv[1])(*sys.argv[2:])
print res

#####
# You can use this simple way.
#client =xmlrpclib.ServerProxy("http://127.0.0.1:5060")
#res = client.htable.listTables()
#print res



## Kamailio - equivalent of routing blocks in Python
##
## KSR - the new dynamic object exporting Kamailio functions
## Router - the old object exporting Kamailio functions
##

## Relevant remarks:
##  * return code -255 is used to propagate the 'exit' behaviour to the
##  parent route block function. The alternative is to use the native
##  Python function sys.exit() (or exit()) -- it throws an exception that
##  is caught by Kamailio and previents the stop of the interpreter.


import sys
import KSR as KSR

# global variables corresponding to defined values (e.g., flags) in kamailio.cfg
FLT_ACC=1
FLT_ACCMISSED=2
FLT_ACCFAILED=3
FLT_NATS=5

FLB_NATB=6
FLB_NATSIPPING=7


# global function to instantiate a kamailio class object
# -- executed when kamailio app_python module is initialized
def mod_init():
    KSR.info("===== from Python mod init\n")
    # dumpObj(KSR)
    return kamailio()


# -- {start defining kamailio class}
class kamailio:
    def __init__(self):
        KSR.info('===== kamailio.__init__\n')


    # executed when kamailio child processes are initialized
    def child_init(self, rank):
        KSR.info('===== kamailio.child_init(%d)\n' % rank)
        return 0


    # SIP request routing
    # -- equivalent of request_route{}
    def ksr_request_route(self, msg):
        # KSR.info("===== request - from kamailio python script\n")
        # KSR.info("===== method [%s] r-uri [%s]\n" % (KSR.pv.get("$rm"),KSR.pv.get("$ru")))

        # per request initial checks
        if self.ksr_route_reqinit(msg)==-255 :
            return 1

        # NAT detection
        if self.ksr_route_natdetect(msg)==-255 :
            return 1

        # CANCEL processing
        if KSR.is_CANCEL() :
            if KSR.tm.t_check_trans()>0 :
                self.ksr_route_relay(msg)
            return 1

        # handle requests within SIP dialogs
        if self.ksr_route_withindlg(msg)==-255 :
            return 1

        # -- only initial requests (no To tag)

        # handle retransmissions
        if KSR.tmx.t_precheck_trans()>0 :
            KSR.tm.t_check_trans()
            return 1

        if KSR.tm.t_check_trans()==0 :
            return 1

        # authentication
        if self.ksr_route_auth(msg)==-255 :
            return 1

        # record routing for dialog forming requests (in case they are routed)
        # - remove preloaded route headers
        KSR.hdr.remove("Route")
        if KSR.is_method_in("IS") :
            KSR.rr.record_route()


        # account only INVITEs
        if KSR.pv.get("$rm")=="INVITE" :
            KSR.setflag(FLT_ACC); # do accounting


        # dispatch requests to foreign domains
        if self.ksr_route_sipout(msg)==-255 :
            return 1

        # # requests for my local domains

        # handle registrations
        if self.ksr_route_registrar(msg)==-255 :
            return 1

        if KSR.corex.has_ruri_user() < 0 :
            # request with no Username in RURI
            KSR.sl.sl_send_reply(484, "Address Incomplete")
            return 1


        # user location service
        self.ksr_route_location(msg)

        return 1


    # wrapper around tm relay function
    def ksr_route_relay(self, msg):
        # enable additional event routes for forwarded requests
        # - serial forking, RTP relaying handling, a.s.o.
        if KSR.is_method_in("IBSU") :
            if KSR.tm.t_is_set("branch_route")<0 :
                KSR.tm.t_on_branch("ksr_branch_manage")

        if KSR.is_method_in("ISU") :
            if KSR.tm.t_is_set("onreply_route")<0 :
                KSR.tm.t_on_reply("ksr_onreply_manage")

        if KSR.is_INVITE() :
            if KSR.tm.t_is_set("failure_route")<0 :
                KSR.tm.t_on_failure("ksr_failure_manage")

        if KSR.tm.t_relay()<0 :
            KSR.sl.sl_reply_error()

        return -255


    # Per SIP request initial checks
    def ksr_route_reqinit(self, msg):
        if not KSR.is_myself(KSR.pv.get("$si")) :
            if not KSR.pv.is_null("$sht(ipban=>$si)") :
                # ip is already blocked
                KSR.dbg("request from blocked IP - " + KSR.pv.get("$rm")
                        + " from " + KSR.pv.get("$fu") + " (IP:"
                        + KSR.pv.get("$si") + ":" + str(KSR.pv.get("$sp")) + ")\n")
                return -255

            if KSR.pike.pike_check_req()<0 :
                KSR.err("ALERT: pike blocking " + KSR.pv.get("$rm")
                        + " from " + KSR.pv.get("$fu") + " (IP:"
                        + KSR.pv.get("$si") + ":" + str(KSR.pv.get("$sp")) + ")\n")
                KSR.pv.seti("$sht(ipban=>$si)", 1)
                return -255

        if KSR.corex.has_user_agent() > 0 :
            ua = KSR.pv.gete("$ua")
            if (ua.find("friendly")!=-1 or ua.find("scanner")!=-1
                    or ua.find("sipcli")!=-1 or ua.find("sipvicious")!=-1
                    or ua.find("VaxSIPUserAgent")!=-1 or ua.find("pplsip")!=-1) :
                KSR.sl.sl_send_reply(200, "Processed")
                return -255

        if KSR.maxfwd.process_maxfwd(10) < 0 :
            KSR.sl.sl_send_reply(483, "Too Many Hops")
            return -255

        if (KSR.is_OPTIONS()
                and KSR.is_myself_ruri()
                and KSR.corex.has_ruri_user() < 0) :
            KSR.sl.sl_send_reply(200, "Keepalive")
            return -255

        if KSR.sanity.sanity_check(17895, 7)<0 :
            KSR.err("Malformed SIP message from "
                    + KSR.pv.get("$si") + ":" + str(KSR.pv.get("$sp")) +"\n")
            return -255


    # Handle requests within SIP dialogs
    def ksr_route_withindlg(self, msg):
        if KSR.siputils.has_totag()<0 :
            return 1

        # sequential request within a dialog should
        # take the path determined by record-routing
        if KSR.rr.loose_route()>0 :
            if self.ksr_route_dlguri(msg)==-255 :
                return -255
            if KSR.is_BYE() :
                # do accounting ...
                KSR.setflag(FLT_ACC)
                # ... even if the transaction fails
                KSR.setflag(FLT_ACCFAILED)
            elif KSR.is_ACK() :
                # ACK is forwarded statelessly
                if self.ksr_route_natmanage(msg)==-255 :
                    return -255
            elif KSR.is_NOTIFY() :
                # Add Record-Route for in-dialog NOTIFY as per RFC 6665.
                KSR.rr.record_route()

            self.ksr_route_relay(msg)
            return -255

        if KSR.is_ACK() :
            if KSR.tm.t_check_trans() >0 :
                # no loose-route, but stateful ACK
                # must be an ACK after a 487
                # or e.g. 404 from upstream server
                self.ksr_route_relay(msg)
                return -255
            else:
                # ACK without matching transaction ... ignore and discard
                return -255

        KSR.sl.sl_send_reply(404, "Not here")
        return -255


    # Handle SIP registrations
    def ksr_route_registrar(self, msg):
        if not KSR.is_REGISTER() :
            return 1
        if KSR.isflagset(FLT_NATS) :
            KSR.setbflag(FLB_NATB)
            # do SIP NAT pinging
            KSR.setbflag(FLB_NATSIPPING)

        if KSR.registrar.save("location", 0)<0 :
            KSR.sl.sl_reply_error()

        return -255


    # User location service
    def ksr_route_location(self, msg):
        rc = KSR.registrar.lookup("location")
        if rc<0 :
            KSR.tm.t_newtran()
            if rc==-1 or rc==-3 :
                KSR.sl.send_reply(404, "Not Found")
                return -255
            elif rc==-2 :
                KSR.sl.send_reply(405, "Method Not Allowed")
                return -255

        # when routing via usrloc, log the missed calls also
        if KSR.is_INVITE() :
            KSR.setflag(FLT_ACCMISSED)

        self.ksr_route_relay(msg)
        return -255



    # IP authorization and user uthentication
    def ksr_route_auth(self, msg):

        if not KSR.is_REGISTER() :
            if KSR.permissions.allow_source_address(1)>0 :
                # source IP allowed
                return 1

        if KSR.is_REGISTER() or KSR.is_myself_furi() :
            # authenticate requests
            if KSR.auth_db.auth_check(KSR.pv.get("$fd"), "subscriber", 1)<0 :
                KSR.auth.auth_challenge(KSR.pv.get("$fd"), 0)
                return -255

            # user authenticated - remove auth header
            if not KSR.is_method_in("RP") :
                KSR.auth.consume_credentials()

        # if caller is not local subscriber, then check if it calls
        # a local destination, otherwise deny, not an open relay here
        if (not KSR.is_myself_furi()) and (not KSR.is_myself_ruri()) :
            KSR.sl.sl_send_reply(403, "Not relaying")
            return -255

        return 1


    # Caller NAT detection
    def ksr_route_natdetect(self, msg):
        KSR.force_rport()
        if KSR.nathelper.nat_uac_test(19)>0 :
            if KSR.is_REGISTER() :
                KSR.nathelper.fix_nated_register()
            elif KSR.siputils.is_first_hop()>0 :
                KSR.nathelper.set_contact_alias()

            KSR.setflag(FLT_NATS)

        return 1


    # RTPProxy control
    def ksr_route_natmanage(self, msg):
        if KSR.siputils.is_request()>0 :
            if KSR.siputils.has_totag()>0 :
                if KSR.rr.check_route_param("nat=yes")>0 :
                    KSR.setbflag(FLB_NATB)

        if (not (KSR.isflagset(FLT_NATS) or KSR.isbflagset(FLB_NATB))) :
            return 1

        KSR.rtpproxy.rtpproxy_manage("co")

        if KSR.siputils.is_request()>0 :
            if not KSR.siputils.has_totag() :
                if KSR.tmx.t_is_branch_route()>0 :
                    KSR.rr.add_rr_param(";nat=yes")

        if KSR.siputils.is_reply()>0 :
            if KSR.isbflagset(FLB_NATB) :
                KSR.nathelper.set_contact_alias()

        return 1


    # URI update for dialog requests
    def ksr_route_dlguri(self, msg):
        if not KSR.isdsturiset() :
            KSR.nathelper.handle_ruri_alias()

        return 1


    # Routing to foreign domains
    def ksr_route_sipout(self, msg):
        if KSR.is_myself_ruri() :
            return 1

        KSR.hdr.append("P-Hint: outbound\r\n")
        self.ksr_route_relay(msg)
        return -255


    # Manage outgoing branches
    # -- equivalent of branch_route[...]{}
    def ksr_branch_manage(self, msg):
        KSR.dbg("new branch ["+ str(KSR.pv.get("$T_branch_idx"))
                    + "] to "+ KSR.pv.get("$ru") + "\n")
        self.ksr_route_natmanage(msg)
        return 1


    # Manage incoming replies
    # -- equivalent of onreply_route[...]{}
    def ksr_onreply_manage(self, msg):
        KSR.dbg("incoming reply\n")
        scode = KSR.pv.get("$rs")
        if scode>100 and scode<299 :
            self.ksr_route_natmanage(msg)

        return 1


    # Manage failure routing cases
    # -- equivalent of failure_route[...]{}
    def ksr_failure_manage(self, msg):
        if self.ksr_route_natmanage(msg)==-255 : return 1

        if KSR.tm.t_is_canceled()>0 :
            return 1

        return 1


    # SIP response handling
    # -- equivalent of reply_route{}
    def ksr_reply_route(self, msg):
        KSR.dbg("response handling - python script\n")

        if KSR.sanity.sanity_check(17604, 6)<0 :
            KSR.err("Malformed SIP response from "
                    + KSR.pv.get("$si") + ":" + str(KSR.pv.get("$sp")) +"\n")
            KSR.set_drop()
            return -255

        return 1


# -- {end defining kamailio class}


# global helper function for debugging purposes
def dumpObj(obj):
    for attr in dir(obj):
        KSR.info("obj.%s = %s\n" % (attr, getattr(obj, attr)))




# -*- coding: utf-8 -*-

"""
    Script for traceback test.
    2012.12.03: Created by: Konstantin M. <evilzluk@gmail.com>
"""

import pprint

class Loggers:

    def __init__(self):
        pass

    def __del__(self):
        pass

    def child_init(self, y):
        return 0

    def BuggyCode_lvl5(self, a):
        a / 0

    def BuggyCode_lvl4(self, a):
        return self.BuggyCode_lvl5(a)

    def BuggyCode_lvl3(self, a):
        return self.BuggyCode_lvl4(a)

    def BuggyCode_lvl2(self, a):
        return self.BuggyCode_lvl3(a)

    def BuggyCode(self, a, b=None):
        return self.BuggyCode_lvl2(a)



def mod_init():
    return Loggers()



if __name__ != "__main__":
    import Router
else:
    mod_init().BuggyCode(0)



import xmlrpclib, httplib, sys

# Usage:  python xmlrpc_test.py command [params...]
#
# python script for sending an xmlrpc command to ser's xmlrpc module.
# Note: it uses a re-defined transport class that does not depend on the
# server side closing the connection (it can be used to send multiple
# commands without closing the connections and it will work without any
# special workarounds in the ser.cfg xmlrpc route).
#
# Credits: the transport class comes from ser_ctl, heavily trimmed to a very
# basic version (the ser_ctl version is much more complex, supports
# authentication, ssl a.s.o). See
# http://git.sip-router.org/cgi-bin/gitweb.cgi?p=ser;a=blob;f=ser_ctl/serctl/serxmlrpc.py#l73 for the original (better) version.
#
#
# History:
# --------
#  2009-07-13  initial version (andrei)
#


XMLRPC_SERVER = "127.0.0.1"
XMLRPC_PORT = 5060

class Transport:
	def __init__(self):
		self.conn=httplib.HTTPConnection(XMLRPC_SERVER, str(XMLRPC_PORT));

	def _http_request(self, uripath, body, host):
		self.conn.request("POST", uripath, body, {})

	def request(self, host, uripath, body, verbose=0):
		self._http_request(uripath, body, host)
		response=self.conn.getresponse()
		if response.status != 200:
			raise xmlrpclib.ProtocolError(host+uripath, response.status,
							response.reason, response.msg)
		data=response.read()
		parser, unmarshaller=xmlrpclib.getparser()
		parser.feed(data)
		parser.close()
		return unmarshaller.close()

if len(sys.argv) < 2:
	sys.exit("Usage: "+sys.argv[0]+" rpc_command [args...]");
transport=Transport()
c=xmlrpclib.ServerProxy("http://" + XMLRPC_SERVER+ ":" + str(XMLRPC_PORT),
						transport)
res=getattr(c, sys.argv[1])(*sys.argv[2:])
print res;



# Python 3 helper program for KEMI
# - print IF conditions with param types for KEMI interpreters
# - print typedefs for functions

PRINTPARAMS=6
# - print mode: typedefs, xtypedefs, common, js, lua, python, pythonparams, ruby, sqlang
PRINTMODE="xtypedefs"
# - two tabs for python params, three for the other cases
# PRINTTABS="\t\t"
PRINTTABS="\t\t\t"
PRINTELSE=""

def printCodeFuncTypedefs(prefix):
	sfunc = "typedef int (*sr_kemi_fm" + prefix + "_f)(sip_msg_t*"
	", str*, str*, str*, str*, str*);"
	for i, c in enumerate(prefix):
		if c == 's':
			sfunc += ", str*"
		else:
			sfunc += ", int"
	sfunc += ");"
	print(sfunc)


def printCodeFuncXTypedefs(prefix):
	sfunc = "typedef sr_kemi_xval_t* (*sr_kemi_xfm" + prefix + "_f)(sip_msg_t*"
	", str*, str*, str*, str*, str*);"
	for i, c in enumerate(prefix):
		if c == 's':
			sfunc += ", str*"
		else:
			sfunc += ", int"
	sfunc += ");"
	print(sfunc)


def printCodeIfEnd(sretfunc):
	print(PRINTTABS + "} else {")
	print(PRINTTABS + "\tLM_ERR(\"invalid parameters for: %.*s\\n\", fname->len, fname->s);")
	print(PRINTTABS + "\treturn " + sretfunc + ";")
	print(PRINTTABS + "}")

def printCodeIfEndCommon(sretfunc):
	print(PRINTTABS + "} else {")
	print(PRINTTABS + "\tLM_ERR(\"invalid parameters for: %.*s\\n\", ket->fname.len, ket->fname.s);")
	print(PRINTTABS + "\treturn " + sretfunc + ";")
	print(PRINTTABS + "}")

def printCodeIfParams(prefix):
	global PRINTELSE
	sparams = ""
	for i, c in enumerate(prefix):
		if i==0:
			if c == 's':
				print(PRINTTABS + PRINTELSE + "if(ket->ptypes[0]==SR_KEMIP_STR")
				sparams += "&vps[" + str(i) +"].s, "
			else:
				print(PRINTTABS + PRINTELSE + "if(ket->ptypes[0]==SR_KEMIP_INT")
				sparams += "vps[" + str(i) +"].n, "
			PRINTELSE = "} else "
		elif i==PRINTPARAMS-1:
			if c == 's':
				print(PRINTTABS + "\t\t&& ket->ptypes[" + str(i) + "]==SR_KEMIP_STR) {")
				sparams += "&vps[" + str(i) +"].s);"
			else:
				print(PRINTTABS + "\t\t&& ket->ptypes[" + str(i) + "]==SR_KEMIP_INT) {")
				sparams += "vps[" + str(i) +"].n);"
		else:
			if c == 's':
				print(PRINTTABS + "\t\t&& ket->ptypes[" + str(i) + "]==SR_KEMIP_STR")
				sparams += "&vps[" + str(i) +"].s, "
			else:
				print(PRINTTABS + "\t\t&& ket->ptypes[" + str(i) + "]==SR_KEMIP_INT")
				sparams += "vps[" + str(i) +"].n, "
	return sparams


def printCodeIfValParams(prefix):
	global PRINTELSE
	sparams = ""
	for i, c in enumerate(prefix):
		if i==0:
			if c == 's':
				print(PRINTTABS + PRINTELSE + "if((ket->ptypes[0] & SR_KEMIP_STR)")
				sparams += "&vps[" + str(i) +"].v.s, "
			else:
				print(PRINTTABS + PRINTELSE + "if((ket->ptypes[0] & SR_KEMIP_INT)")
				sparams += "vps[" + str(i) +"].v.n, "
			PRINTELSE = "} else "
		elif i==PRINTPARAMS-1:
			if c == 's':
				print(PRINTTABS + "\t\t&& (ket->ptypes[" + str(i) + "] & SR_KEMIP_STR)) {")
				sparams += "&vps[" + str(i) +"].v.s);"
			else:
				print(PRINTTABS + "\t\t&& (ket->ptypes[" + str(i) + "] & SR_KEMIP_INT)) {")
				sparams += "vps[" + str(i) +"].v.n);"
		else:
			if c == 's':
				print(PRINTTABS + "\t\t&& (ket->ptypes[" + str(i) + "] & SR_KEMIP_STR)")
				sparams += "&vps[" + str(i) +"].v.s, "
			else:
				print(PRINTTABS + "\t\t&& (ket->ptypes[" + str(i) + "] & SR_KEMIP_INT)")
				sparams += "vps[" + str(i) +"].v.n, "
	return sparams


def printCodeIfJS(prefix):
	global PRINTELSE
	sparams = printCodeIfParams(prefix)
	print("\t\t\t\tif(ket->rtype==SR_KEMIP_XVAL) {")
	sfunc = PRINTTABS + "\t\txret = ((sr_kemi_xfm" + prefix + "_f)(ket->func))(env_J->msg,\n" + PRINTTABS + "\t\t\t"
	print(sfunc + sparams)
	print(PRINTTABS + "\t\treturn sr_kemi_jsdt_return_xval(J, ket, xret);")
	print("\t\t\t\t} else {")
	sfunc = PRINTTABS + "\t\tret = ((sr_kemi_fm" + prefix + "_f)(ket->func))(env_J->msg,\n" + PRINTTABS + "\t\t\t"
	print(sfunc + sparams)
	print(PRINTTABS + "\t\treturn sr_kemi_jsdt_return_int(J, ket, ret);")
	print("\t\t\t\t}")


def printCodeIfLua(prefix):
	global PRINTELSE
	sparams = printCodeIfParams(prefix)
	print("\t\t\t\tif(ket->rtype==SR_KEMIP_XVAL) {")
	sfunc = PRINTTABS + "\t\txret = ((sr_kemi_xfm" + prefix + "_f)(ket->func))(env_L->msg,\n" + PRINTTABS + "\t\t\t"
	print(sfunc + sparams)
	print(PRINTTABS + "\t\treturn sr_kemi_lua_return_xval(L, ket, xret);")
	print("\t\t\t\t} else {")
	sfunc = PRINTTABS + "\t\tret = ((sr_kemi_fm" + prefix + "_f)(ket->func))(env_L->msg,\n" + PRINTTABS + "\t\t\t"
	print(sfunc + sparams)
	print(PRINTTABS + "\t\treturn sr_kemi_lua_return_int(L, ket, ret);")
	print("\t\t\t\t}")


def printCodeIfPython(prefix):
	global PRINTELSE
	sparams = printCodeIfParams(prefix)
	print("\t\t\t\tif(ket->rtype==SR_KEMIP_XVAL) {")
	sfunc = PRINTTABS + "\t\txret = ((sr_kemi_xfm" + prefix + "_f)(ket->func))(lmsg,\n" + PRINTTABS + "\t\t\t"
	print(sfunc + sparams)
	print(PRINTTABS + "\t\treturn sr_kemi_apy_return_xval(ket, xret);")
	print("\t\t\t\t} else {")
	sfunc = PRINTTABS + "\t\tret = ((sr_kemi_fm" + prefix + "_f)(ket->func))(lmsg,\n" + PRINTTABS + "\t\t\t"
	print(sfunc + sparams)
	print(PRINTTABS + "\t\treturn sr_kemi_apy_return_int(ket, ret);")
	print("\t\t\t\t}")


def printCodeIfPythonParams(prefix):
	global PRINTELSE
	sfunc = PRINTTABS + "\tif(!PyArg_ParseTuple(args, \"" + prefix + ":kemi-param-" + prefix + "\",\n" + PRINTTABS + "\t\t\t"
	slen = ""
	sdbg = PRINTTABS + "\tLM_DBG(\"params[%d] for: %.*s are:"
	sval = ""
	for i, c in enumerate(prefix):
		if i==0:
			if c == 's':
				print(PRINTTABS + PRINTELSE + "if(ket->ptypes[0]==SR_KEMIP_STR")
				sfunc += "&vps[" + str(i) +"].s.s, "
				slen += PRINTTABS + "\tvps[" + str(i) + "].s.len = strlen(vps[" + str(i) + "].s.s);\n"
				sdbg += " [%s]"
				sval += "vps[" + str(i) + "].s.s, "
			else:
				print(PRINTTABS + PRINTELSE + "if(ket->ptypes[0]==SR_KEMIP_INT")
				sfunc += "&vps[" + str(i) +"].n, "
				sdbg += " [%d]"
				sval += "vps[" + str(i) + "].n, "
			PRINTELSE = "} else "
		elif i==PRINTPARAMS-1:
			if c == 's':
				print(PRINTTABS + "\t\t&& ket->ptypes[" + str(i) + "]==SR_KEMIP_STR) {")
				sfunc += "&vps[" + str(i) +"].s.s)) {"
				slen += PRINTTABS + "\tvps[" + str(i) + "].s.len = strlen(vps[" + str(i) + "].s.s);\n"
				sdbg += " [%s]"
				sval += "vps[" + str(i) + "].s.s"
			else:
				print(PRINTTABS + "\t\t&& ket->ptypes[" + str(i) + "]==SR_KEMIP_INT) {")
				sfunc += "&vps[" + str(i) +"].n)) {"
				sdbg += " [%d]"
				sval += "vps[" + str(i) + "].n"
		else:
			if c == 's':
				print(PRINTTABS + "\t\t&& ket->ptypes[" + str(i) + "]==SR_KEMIP_STR")
				sfunc += "&vps[" + str(i) +"].s.s, "
				slen += PRINTTABS + "\tvps[" + str(i) + "].s.len = strlen(vps[" + str(i) + "].s.s);\n"
				sdbg += " [%s]"
				sval += "vps[" + str(i) + "].s.s, "
			else:
				print(PRINTTABS + "\t\t&& ket->ptypes[" + str(i) + "]==SR_KEMIP_INT")
				sfunc += "&vps[" + str(i) +"].n, "
				sdbg += " [%d]"
				sval += "vps[" + str(i) + "].n, "
	print(sfunc)
	print(PRINTTABS + "\t\tLM_ERR(\"unable to retrieve " + prefix + " params %d\\n\", i);")
	print(PRINTTABS + "\t\treturn sr_kemi_apy_return_false();")
	print(PRINTTABS + "\t}")
	print(slen)
	print(sdbg + "\\n\",\n" + PRINTTABS + "\t\t\t" + "i, fname.len, fname.s,")
	print(PRINTTABS + "\t\t\t" + sval + ");")


def printCodeIfRuby(prefix):
	global PRINTELSE
	sparams = printCodeIfParams(prefix)
	print("\t\t\t\tif(ket->rtype==SR_KEMIP_XVAL) {")
	sfunc = PRINTTABS + "\t\txret = ((sr_kemi_xfm" + prefix + "_f)(ket->func))(env_R->msg,\n" + PRINTTABS + "\t\t\t"
	print(sfunc + sparams)
	print(PRINTTABS + "\t\treturn sr_kemi_ruby_return_xval(ket, xret);")
	print("\t\t\t\t} else {")
	sfunc = PRINTTABS + "\t\tret = ((sr_kemi_fm" + prefix + "_f)(ket->func))(env_R->msg,\n" + PRINTTABS + "\t\t\t"
	print(sfunc + sparams)
	print(PRINTTABS + "\t\treturn sr_kemi_ruby_return_int(ket, ret);")
	print("\t\t\t\t}")


def printCodeIfSQLang(prefix):
	global PRINTELSE
	sparams = printCodeIfParams(prefix)
	print("\t\t\t\tif(ket->rtype==SR_KEMIP_XVAL) {")
	sfunc = PRINTTABS + "\t\txret = ((sr_kemi_xfm" + prefix + "_f)(ket->func))(env_J->msg,\n" + PRINTTABS + "\t\t\t"
	print(sfunc + sparams)
	print(PRINTTABS + "\t\treturn sr_kemi_sqlang_return_xval(J, ket, xret);")
	print("\t\t\t\t} else {")
	sfunc = PRINTTABS + "\t\tret = ((sr_kemi_fm" + prefix + "_f)(ket->func))(env_J->msg,\n" + PRINTTABS + "\t\t\t"
	print(sfunc + sparams)
	print(PRINTTABS + "\t\treturn sr_kemi_sqlang_return_int(J, ket, ret);")
	print("\t\t\t\t}")


def printCodeIfCommon(prefix):
	global PRINTELSE
	sparams = printCodeIfValParams(prefix)
	print("\t\t\t\tif(ket->rtype & SR_KEMIP_XVAL) {")
	sfunc = PRINTTABS + "\t\treturn ((sr_kemi_xfm" + prefix + "_f)(ket->func))(msg,\n" + PRINTTABS + "\t\t\t"
	print(sfunc + sparams)
	print("\t\t\t\t} else {")
	sfunc = PRINTTABS + "\t\tret = ((sr_kemi_fm" + prefix + "_f)(ket->func))(msg,\n" + PRINTTABS + "\t\t\t"
	print(sfunc + sparams)
	print(PRINTTABS + "\t\treturn sr_kemi_return_int(ket, ret);")
	print("\t\t\t\t}")


# generated possible strings of length k with chars from set.
def printAllKLength(cset, k):

	n = len(cset)
	printAllKLengthRec(cset, "", n, k)


def printAllKLengthRec(cset, prefix, n, k):

	if (k == 0) :
		# print(prefix)
		if PRINTMODE == "js":
			printCodeIfJS(prefix)
		elif PRINTMODE == "lua":
			printCodeIfLua(prefix)
		elif PRINTMODE == "python":
			printCodeIfPython(prefix)
		elif PRINTMODE == "pythonparams":
			printCodeIfPythonParams(prefix)
		elif PRINTMODE == "ruby":
			printCodeIfRuby(prefix)
		elif PRINTMODE == "sqlang":
			printCodeIfSQLang(prefix)
		elif PRINTMODE == "common":
			printCodeIfCommon(prefix)
		elif PRINTMODE == "xtypedefs":
			printCodeFuncXTypedefs(prefix)
		else:
			printCodeFuncTypedefs(prefix)
		return

	for i in range(n):
		newPrefix = prefix + cset[i]
		printAllKLengthRec(cset, newPrefix, n, k - 1)


# main statements
if __name__ == "__main__":

	charset = ['s', 'n']
	k = PRINTPARAMS
	printAllKLength(charset, k)
	if PRINTMODE == "js":
		printCodeIfEnd("app_jsdt_return_false(J)")
	elif PRINTMODE == "lua":
		printCodeIfEnd("app_lua_return_false(L)")
	elif PRINTMODE == "python":
		printCodeIfEnd("sr_kemi_apy_return_false()")
	elif PRINTMODE == "pythonparams":
		printCodeIfEnd("sr_kemi_apy_return_false()")
	elif PRINTMODE == "ruby":
		printCodeIfEnd("Qfalse")
	elif PRINTMODE == "sqlang":
		printCodeIfEnd("app_sqlang_return_false(J)")
	elif PRINTMODE == "common":
		printCodeIfEndCommon("sr_kemi_return_false(ket)")





#!/usr/bin/python3

import KSR

# this assumes you have your mock KSR.py in the local directory e.g. /test/
# and your kemi code is in ../conf/kamailio.py

sys.path.insert(0, "../conf/")
import kamailio as kamailio


#return sip:hello@world only if $ru is passed to pv.get
KSR._mock_data['pv']['get'] = {}
KSR._mock_data['pv']['get']['$ru'] = "sip:hello@world"
print("Got a value of: " + KSR.pv.get("$ru"))

#return maxfwd.process_maxfwd return 2 regardless of value passed
KSR._mock_data['maxfwd']['process_maxfwd'] = 2
KSR.maxfwd.process_maxfwd(10)
print("Got a value of: " + str(KSR.maxfwd.process_maxfwd(10)))

#set a function pointer to see if hdr.append is called
appendCalled = False
def appendHeader(param0: str):
    global appendCalled
    if param0.startswith("X-HDR:"):
        appendCalled = True
    return 1

KSR._mock_data['hdr']['append'] = appendHeader
KSR.hdr.append("X-HDR: my-header")

k = kamailio.kamailio()
k.ksr_request_route(None) # Call the kemi script, the mock implementations will be called

# Validate the results
if appendCalled:
    print("hdr.append successfully called!")
else:
    print("hdr.append failed to be called")



#! /usr/bin/env python3
"""
 * $Id$
 *
 * Copyright (C) 2006 iptelorg GmbH
 *
 * This file is part of ser, a free SIP server.
 *
 * ser is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version
 *
 * For a license to use the ser software under conditions
 * other than those described here, or to purchase support for this
 * software, please contact iptel.org by e-mail at the following addresses:
 *    info@iptel.org
 *
 * ser is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 *
"""

import sys, re

max_depth = 120
debug = 0

re_main_route = re.compile("^([a-z]+_)*route[\s\t]*(?![\(\)])[\s\t]*\{?", re.I)  # Regular expression to match main routes
re_def_route = re.compile("^([a-z]+_)*route ?(\[\"?([A-Za-z0-9-_:]+)\"?\])+[\s\t]*\{?", re.I)  # Regular expression to match defined routes
re_call_route = re.compile("^(.*\([\s\t!]*)?route\(\"?([A-Za-z0-9-_]+)\"?\)", re.I)  # Regular expression to match called routes

routes = {}  # Dictionary to store main routes
f_routes = {}  # Dictionary to store failure routes
b_routes = {}  # Dictionary to store branch routes
or_routes = {}  # Dictionary to store onreply routes
r_routes = {}  # Dictionary to store reply routes
s_routes = {}  # Dictionary to store onsend routes
e_routes = {}  # Dictionary to store event routes

def log(_s):
    if debug:
        print(_s)

def print_route_level(_l, _n):
    log(f"prl: {_l}, {_n}")
    if _l > max_depth:
        return
    spacer = ""
    route = ""
    for i in range(_l):
        spacer += "|  "
        if i < _l - 1:
            route += "|  "
        else:
            route += "\- " + str(_n)
    if len(spacer) > 0:
        print(spacer)
    if len(route) > 0:
        print(route)

def traverse_routes(_level, _name):
    log(f"tr: {_level}, {_name}")
    if _level > max_depth:
        print("warning: max_depth reached")
        return
    print_route_level(_level, _name)
    if _name in routes:
        for r in routes[_name]:
            traverse_routes(_level + 1, r)

# Checking command-line arguments
if len(sys.argv) < 2:
    raise Exception('wrong number of arguments\nusage: ' + sys.argv[0] + ' configuration-file [max_depth]')
if len(sys.argv) == 3:
    max_depth = int(sys.argv[2])
cfg = open(sys.argv[1], "r")
if cfg is None:
    raise Exception('Missing config file')

# Reading the configuration file
line = cfg.readline()
rt = routes
while line:
    line = line.strip()
    if not line.startswith("#"):
        log(line)
        main_match = re_main_route.search(line)
        def_match = re_def_route.search(line)
        call_match = re_call_route.search(line)

        # Matching a called route
        if call_match is not None:
            log("CALL: " + line)
            name = call_match.group(2)
            log(f"{rname}:{name}")
            rt[rname].append(name)

        # Matching a defined route
        elif def_match is not None:
            log("DEF: " + line)
            rtype = def_match.group(1)
            rname = def_match.group(3)
            if rtype == "failure_":
                rt = f_routes
                if rname is None:
                    rname = "failure"
            elif rtype == "onreply_":
                rt = or_routes
                if rname is None:
                    rname = "onreply"
            elif rtype == "reply_":
                rt = r_routes
                if rname is None:
                    rname = "reply"
            elif rtype == "onsend_":
                rt = s_routes
                if rname is None:
                    rname = "onsend"
            elif rtype == "branch_":
                rt = b_routes
                if rname is None:
                    rname = "branch"
            elif rtype == "event_":
                rt = e_routes
                if rname is None:
                    rname = "event"
            else:
                rt = routes
            log(rname)
            rt[rname] = []

        # Matching a main route
        elif main_match is not None:
            log("MAIN: " + line)
            rtype = main_match.group(1)
            if rtype == "failure_":
                rt = f_routes
                rname = "failure"
            elif rtype == "onreply_":
                rt = or_routes
                rname = "onreply"
            elif rtype == "reply_":
                rt = r_routes
                rname = "reply"
            elif rtype == "onsend_":
                rt = s_routes
                rname = "onsend"
            elif rtype == "branch_":
                rt = b_routes
                rname = "branch"
            elif rtype == "event_":
                rt = e_routes
                rname = "event"
            else:
                rt = routes
                rname = "Main"
            log(rname)
            rt[rname] = []

    line = cfg.readline()

log(f"routes: {routes}")
log(f"branch_routes: {b_routes}")
log(f"failure_routes: {f_routes}")
log(f"onreply_routes: {or_routes}")
log(f"reply_routes: {r_routes}")
log(f"onsend_routes: {s_routes}")
log(f"event_routes: {e_routes}")

# Checking for missing routes
for name in routes.keys():
    for val in routes[name]:
        if val not in routes:
            print(f"Missing Route {val}!!!")

print("\nMain")
traverse_routes(0, "Main")

# Printing branch routes
if len(b_routes) > 0:
    print("\nBranch routes\n-------------")
    for br in b_routes.keys():
        print(f"\n{br}")
        for r in b_routes[br]:
            traverse_routes(1, r)

# Printing onsend routes
if len(s_routes) > 0:
    print("\nSend routes\n-----------")
    for sr in s_routes.keys():
        print(f"\n{sr}")
        for r in s_routes[sr]:
            traverse_routes(1, r)

# Printing failure routes
if len(f_routes) > 0:
    print("\nFailure routes\n--------------")
    for fr in f_routes.keys():
        print(f"\n{fr}")
        for r in f_routes[fr]:
            traverse_routes(1, r)

# Printing onreply routes
if len(or_routes) > 0:
    print("\nOnreply routes\n--------------")
    for onr in or_routes.keys():
        print(f"\n{onr}")
        for r in or_routes[onr]:
            traverse_routes(1, r)

# Printing reply routes
if len(r_routes) > 0:
    print("\nReply routes\n--------------")
    for onr in r_routes.keys():
        print(f"\n{onr}")
        for r in r_routes[onr]:
            traverse_routes(1, r)

# Printing event routes
if len(e_routes) > 0:
    print("\nEvent routes\n--------------")
    for onr in e_routes.keys():
        print(f"\n{onr}")
        for r in e_routes[onr]:
            traverse_routes(1, r)

print()
cfg.close()



## Kamailio - equivalent of routing blocks in Python
##
## KSR - the new dynamic object exporting Kamailio functions
## Router - the old object exporting Kamailio functions
##

## Relevant remarks:
##  * return code -255 is used to propagate the 'exit' behaviour to the
##  parent route block function. The alternative is to use the native
##  Python function sys.exit() (or exit()) -- it throws an exception that
##  is caught by Kamailio and prevents the stop of the interpreter.


import sys
import KSR as KSR

# global variables corresponding to defined values (e.g., flags) in kamailio.cfg
FLT_ACC=1
FLT_ACCMISSED=2
FLT_ACCFAILED=3
FLT_NATS=5

FLB_NATB=6
FLB_NATSIPPING=7


# SIP request routing
# -- equivalent of request_route{}
def ksr_request_route():
    # KSR.info("===== request - from kamailio python script\n")
    # KSR.info("===== method [%s] r-uri [%s]\n" % (KSR.pv.get("$rm"),KSR.pv.get("$ru")))

    # per request initial checks
    if ksr_route_reqinit()==-255 :
        return 1

    # NAT detection
    if ksr_route_natdetect()==-255 :
        return 1

    # CANCEL processing
    if KSR.is_CANCEL() :
        if KSR.tm.t_check_trans()>0 :
            ksr_route_relay()
        return 1

    # handle retransmissions
    if not KSR.is_ACK() :
        if KSR.tmx.t_precheck_trans()>0 :
            KSR.tm.t_check_trans()
            return 1

        if KSR.tm.t_check_trans()==0 :
            return 1

    # handle requests within SIP dialogs
    if ksr_route_withindlg()==-255 :
        return 1

    # -- only initial requests (no To tag)

    # authentication
    if ksr_route_auth()==-255 :
        return 1

    # record routing for dialog forming requests (in case they are routed)
    # - remove preloaded route headers
    KSR.hdr.remove("Route")
    if KSR.is_method_in("IS") :
        KSR.rr.record_route()


    # account only INVITEs
    if KSR.is_INVITE() :
        KSR.setflag(FLT_ACC); # do accounting


    # dispatch requests to foreign domains
    if ksr_route_sipout()==-255 :
        return 1

    # # requests for my local domains

    # handle registrations
    if ksr_route_registrar()==-255 :
        return 1

    if KSR.corex.has_ruri_user() < 0 :
        # request with no Username in RURI
        KSR.sl.sl_send_reply(484, "Address Incomplete")
        return 1


    # user location service
    ksr_route_location()

    return 1


# wrapper around tm relay function
def ksr_route_relay():
    # enable additional event routes for forwarded requests
    # - serial forking, RTP relaying handling, a.s.o.
    if KSR.is_method_in("IBSU") :
        if KSR.tm.t_is_set("branch_route")<0 :
            KSR.tm.t_on_branch("ksr_branch_manage")

    if KSR.is_method_in("ISU") :
        if KSR.tm.t_is_set("onreply_route")<0 :
            KSR.tm.t_on_reply("ksr_onreply_manage")

    if KSR.is_INVITE() :
        if KSR.tm.t_is_set("failure_route")<0 :
            KSR.tm.t_on_failure("ksr_failure_manage")

    if KSR.tm.t_relay()<0 :
        KSR.sl.sl_reply_error()

    return -255


# Per SIP request initial checks
def ksr_route_reqinit():
    # no connect for sending replies
    KSR.set_reply_no_connect()
    # enforce symmetric signaling
    # send back replies to the source address of request
    KSR.force_rport()

    if not KSR.is_myself_srcip() :
        srcip = KSR.kx.get_srcip()
        if KSR.htable.sht_match_name("ipban", "eq", srcip) > 0 :
            # ip is already blocked
            KSR.dbg("request from blocked IP - " + KSR.kx.get_method()
                    + " from " + KSR.kx.get_furi() + " (IP:"
                    + srcip + ":" + str(KSR.pv.get("$sp")) + ")\n")
            return -255

        if KSR.pike.pike_check_req()<0 :
            KSR.err("ALERT: pike blocking " + KSR.kx.get_method()
                    + " from " + KSR.kx.get_furi() + " (IP:"
                    + srcip + ":" + str(KSR.pv.get("$sp")) + ")\n")
            KSR.htable.sht_seti("ipban", srcip, 1)
            return -255

    if KSR.corex.has_user_agent() > 0 :
        ua = KSR.kx.gete_ua()
        if (ua.find("friendly")!=-1 or ua.find("scanner")!=-1
                or ua.find("sipcli")!=-1 or ua.find("sipvicious")!=-1
                or ua.find("VaxSIPUserAgent")!=-1 or ua.find("pplsip")!=-1) :
            KSR.sl.sl_send_reply(200, "Processed")
            return -255

    if KSR.maxfwd.process_maxfwd(10) < 0 :
        KSR.sl.sl_send_reply(483, "Too Many Hops")
        return -255

    if (KSR.is_OPTIONS()
            and KSR.is_myself_ruri()
            and KSR.corex.has_ruri_user() < 0) :
        KSR.sl.sl_send_reply(200, "Keepalive")
        return -255

    if KSR.sanity.sanity_check(17895, 7)<0 :
        KSR.err("Malformed SIP message from "
                + KSR.pv.get("$si") + ":" + str(KSR.pv.get("$sp")) +"\n")
        return -255


# Handle requests within SIP dialogs
def ksr_route_withindlg():
    if KSR.siputils.has_totag()<0 :
        return 1

    # sequential request withing a dialog should
    # take the path determined by record-routing
    if KSR.rr.loose_route()>0 :
        if ksr_route_dlguri()==-255 :
            return -255
        if KSR.is_BYE() :
            # do accounting ...
            KSR.setflag(FLT_ACC)
            # ... even if the transaction fails
            KSR.setflag(FLT_ACCFAILED)
        elif KSR.is_ACK() :
            # ACK is forwarded statelessly
            if ksr_route_natmanage()==-255 :
                return -255
        elif KSR.is_NOTIFY() :
            # Add Record-Route for in-dialog NOTIFY as per RFC 6665.
            KSR.rr.record_route()

        ksr_route_relay()
        return -255

    if KSR.is_ACK() :
        if KSR.tm.t_check_trans() >0 :
            # no loose-route, but stateful ACK
            # must be an ACK after a 487
            # or e.g. 404 from upstream server
            ksr_route_relay()
            return -255
        else:
            # ACK without matching transaction ... ignore and discard
            return -255

    KSR.sl.sl_send_reply(404, "Not here")
    return -255


# Handle SIP registrations
def ksr_route_registrar():
    if not KSR.is_REGISTER() :
        return 1
    if KSR.isflagset(FLT_NATS) :
        KSR.setbflag(FLB_NATB)
        # do SIP NAT pinging
        KSR.setbflag(FLB_NATSIPPING)

    if KSR.registrar.save("location", 0)<0 :
        KSR.sl.sl_reply_error()

    return -255


# User location service
def ksr_route_location():
    rc = KSR.registrar.lookup("location")
    if rc<0 :
        KSR.tm.t_newtran()
        if rc==-1 or rc==-3 :
            KSR.sl.send_reply(404, "Not Found")
            return -255
        elif rc==-2 :
            KSR.sl.send_reply(405, "Method Not Allowed")
            return -255

    # when routing via usrloc, log the missed calls also
    if KSR.is_INVITE() :
        KSR.setflag(FLT_ACCMISSED)

    ksr_route_relay()
    return -255



# IP authorization and user uthentication
def ksr_route_auth():

    if not KSR.is_REGISTER() :
        if KSR.permissions.allow_source_address(1)>0 :
            # source IP allowed
            return 1

    if KSR.is_REGISTER() or KSR.is_myself_furi() :
        # authenticate requests
        if KSR.auth_db.auth_check(KSR.pv.get("$fd"), "subscriber", 1)<0 :
            KSR.auth.auth_challenge(KSR.pv.get("$fd"), 0)
            return -255

        # user authenticated - remove auth header
        if not KSR.is_method_in("RP") :
            KSR.auth.consume_credentials()

    # if caller is not local subscriber, then check if it calls
    # a local destination, otherwise deny, not an open relay here
    if (not KSR.is_myself_furi()) and (not KSR.is_myself_ruri()) :
        KSR.sl.sl_send_reply(403, "Not relaying")
        return -255

    return 1


# Caller NAT detection
def ksr_route_natdetect():
    KSR.force_rport()
    if KSR.nathelper.nat_uac_test(19)>0 :
        if KSR.is_REGISTER() :
            KSR.nathelper.fix_nated_register()
        elif KSR.siputils.is_first_hop()>0 :
            KSR.nathelper.set_contact_alias()

        KSR.setflag(FLT_NATS)

    return 1


# RTPProxy control
def ksr_route_natmanage():
    if KSR.siputils.is_request()>0 :
        if KSR.siputils.has_totag()>0 :
            if KSR.rr.check_route_param("nat=yes")>0 :
                KSR.setbflag(FLB_NATB)

    if (not (KSR.isflagset(FLT_NATS) or KSR.isbflagset(FLB_NATB))) :
        return 1

    KSR.rtpproxy.rtpproxy_manage("co")

    if KSR.siputils.is_request()>0 :
        if not KSR.siputils.has_totag() :
            if KSR.tmx.t_is_branch_route()>0 :
                KSR.rr.add_rr_param(";nat=yes")

    if KSR.siputils.is_reply()>0 :
        if KSR.isbflagset(FLB_NATB) :
            KSR.nathelper.set_contact_alias()

    return 1


# URI update for dialog requests
def ksr_route_dlguri():
    if not KSR.isdsturiset() :
        KSR.nathelper.handle_ruri_alias()

    return 1


# Routing to foreign domains
def ksr_route_sipout():
    if KSR.is_myself_ruri() :
        return 1

    KSR.hdr.append("P-Hint: outbound\r\n")
    ksr_route_relay()
    return -255


# Manage outgoing branches
# -- equivalent of branch_route[...]{}
def ksr_branch_manage():
    KSR.dbg("new branch ["+ str(KSR.pv.get("$T_branch_idx"))
                + "] to "+ KSR.pv.get("$ru") + "\n")
    ksr_route_natmanage()
    return 1


# Manage incoming replies
# -- equivalent of onreply_route[...]{}
def ksr_onreply_manage():
    KSR.dbg("incoming reply\n")
    scode = KSR.pv.get("$rs")
    if scode>100 and scode<299 :
        ksr_route_natmanage()

    return 1


# Manage failure routing cases
# -- equivalent of failure_route[...]{}
def ksr_failure_manage():
    if ksr_route_natmanage()==-255 : return 1

    if KSR.tm.t_is_canceled()>0 :
        return 1

    return 1


# SIP response handling
# -- equivalent of reply_route{}
def ksr_reply_route():
    KSR.dbg("response handling - python script\n")

    if KSR.sanity.sanity_check(17604, 6)<0 :
        KSR.err("Malformed SIP response from "
                + KSR.pv.get("$si") + ":" + str(KSR.pv.get("$sp")) +"\n")
        KSR.set_drop()
        return -255

    return 1




# global helper function for debugging purposes
def dumpObj(obj):
    for attr in dir(obj):
        KSR.info("obj.%s = %s\n" % (attr, getattr(obj, attr)))




# -*- coding: utf-8 -*-

"""
    Script for logging test.
    Added:   2012-12-03: Created by: Konstantin M. <evilzluk@gmail.com>
    Changed: 2012-12-11: Changed code to use Router.Logger
    Changed: 2012-12-11: Changed code to use Router.Ranks
"""

import Router.Logger as Logger
import Router.Ranks as Ranks

"""
    Router.Logger Module Properties:
       Log Levels:
           L_ALERT
           L_BUG
           L_CRIT2
           L_CRIT
           L_ERR
           L_WARN
           L_NOTICE
           L_INFO
           L_DBG

       Log Facilities:
           DEFAULT_FACILITY

    Router.Logger Module Methods:
       LM_GEN1(self, int log_level, str msg)
       LM_GEN2(self, int log_facility, int log_level, str msg)
       LM_ALERT(self, str msg)
       LM_CRIT(self, str msg)
       LM_ERR(self, str msg)
       LM_WARN(self, str msg)
       LM_NOTICE(self, str msg)
       LM_INFO(self, str msg)
       LM_DBG(self, str msg)


    Router.Ranks Module Properties:
        PROC_MAIN
        PROC_TIMER
        PROC_RPC
        PROC_FIFO
        PROC_TCP_MAIN
        PROC_UNIXSOCK
        PROC_ATTENDANT
	PROC_INIT
	PROC_NOCHLDINIT
	PROC_SIPINIT
	PROC_SIPRPC
	PROC_MIN

"""

class Loggers:

    def __init__(self):
	pass

    def __del__(self):
	pass

    def child_init(self, rank):
	if rank == Ranks.PROC_MAIN:
	    Logger.LM_ERR("rank is PROC_MAIN")
	elif rank == Ranks.PROC_TIMER:
	    Logger.LM_ERR("rank is PROC_TIMER")
	elif rank == Ranks.PROC_RPC:
	    Logger.LM_ERR("rank is PROC_RPC")
	elif rank == Ranks.PROC_FIFO:
	    Logger.LM_ERR("rank is PROC_FIFO")
	elif rank == Ranks.PROC_TCP_MAIN:
	    Logger.LM_ERR("rank is PROC_TCP_MAIN")
	elif rank == Ranks.PROC_UNIXSOCK:
	    Logger.LM_ERR("rank is PROC_UNIXSOCK")
	elif rank == Ranks.PROC_ATTENDANT:
	    Logger.LM_ERR("rank is PROC_ATTENDANT")
	elif rank == Ranks.PROC_INIT:
	    Logger.LM_ERR("rank is PROC_INIT")
	elif rank == Ranks.PROC_NOCHLDINIT:
	    Logger.LM_ERR("rank is PROC_NOCHLDINIT")
	elif rank == Ranks.PROC_SIPINIT:
	    Logger.LM_ERR("rank is PROC_SIPINIT")
	elif rank == Ranks.PROC_SIPRPC:
	    Logger.LM_ERR("rank is PROC_SIPRPC")
	elif rank == Ranks.PROC_MIN:
	    Logger.LM_ERR("rank is PROC_MIN")

	return 0

    def TestLoggers(self, msg, args):
	Logger.LM_GEN1(Logger.L_INFO,                           	"Loggers.py:     LM_GEN1: msg: %s" % str(args))
	Logger.LM_GEN2(Logger.L_INFO, Logger.DEFAULT_FACILITY,	        "Loggers.py:     LM_GEN2: msg: %s" % str(args))
	Logger.LM_ALERT(                                        	"Loggers.py:    LM_ALERT: msg: %s" % str(args))
	Logger.LM_CRIT(                                         	"Loggers.py:     LM_CRIT: msg: %s" % str(args))
	Logger.LM_ERR(                                          	"Loggers.py:      LM_ERR: msg: %s" % str(args))
	Logger.LM_WARN(                                         	"Loggers.py:     LM_WARN: msg: %s" % str(args))
	Logger.LM_NOTICE(                                       	"Loggers.py:   LM_NOTICE: msg: %s" % str(args))
	Logger.LM_INFO(                                         	"Loggers.py:     LM_INFO: msg: %s" % str(args))
	Logger.LM_DBG(                                          	"Loggers.py:      LM_DBG: msg: %s" % str(args))
	return 1

def mod_init():
    return Loggers()




# tool to generate the modules.md content
#

import os, json, sys, time, fnmatch, re, importlib

PATH_GENERATED_OBJ = "KSR.luacheckrc"

class ModuleObjGenerator(object):

    # Contains the output until it should be written to disk
    object_string = ""

    def execute(self, data):
        # Validate that we got some methods back. 155 is an arbitrary large number.
        if len(data) < 1:
            print("ERR: Invalid data")
            exit()

        functions_parsed = self.parse_function_list(data)
        self.output_obj(functions_parsed)

        print ("File created successfully at " + PATH_GENERATED_OBJ)

    def parse_function_list(self, functions):
        data = {}

        for elem in functions:
            module = elem["module"]

            # TODO: What about the hdr, pv, x sub-module?
            if module == "":
                module = "_core"

            if module not in data:
                data[module] = []

            data[module].append({"name": elem["name"], "return": elem["ret"], "params": elem["params"]})

        return data

    def output_obj(self, data):

        for key in sorted(data):
            methods = data[key]
            # Sort the functions by name alphabetically
            methods = sorted(methods, key = lambda k: k['name'])
            self.object_module_content(key, methods)

        self.obj_write()

        return True

    def object_module_content(self, module, methods):
        if module != "_core":
            self.object_string += "\t\t\t\t" + module + " = {\n\t\t\t\t\tfields= {\n"

        for value in methods:
            if module != "_core":
                self.object_string += "\t\t\t\t\t\t" + value["name"] + " = {},\n"
            else:
                self.object_string += "\t\t\t\t" + value["name"] + " = {},\n"

        if module != "_core":
            self.object_string += "\t\t\t\t\t},\n\t\t\t\t},\n"
        return True

    def obj_write(self):
        f = open(PATH_GENERATED_OBJ, "w")
        f.write("stds.KSR = {\n\tread_globals = {\n\t\tKSR = {\n\t\t\tfields = {\n")
        f.write(self.object_string)
        f.write("\t\t\t\tx = {\n\t\t\t\t\tfields= {\n")
        f.write("\t\t\t\t\t\tdrop= {},\n")
        f.write("\t\t\t\t\t\texit= {},\n")
        f.write("\t\t\t\t\t\tmodf= {},\n")
        f.write("\t\t\t\t\t},\n\t\t\t\t},\n")
        f.write("\t\t\t}\n\t\t}\n\t}\n}\n")
        f.write("\nstd = \"min+KSR\"\n")
        f.close()
        return True

    def read_file_to_string(self, filename):
        path = self.PATH_MODULES_DOCS + filename
        if os.path.isfile(path):
            with open(path, 'r') as myfile:
                return myfile.read() + "\n"
        return ""


class KemiFileExportParser(object):
    # These functions are created by a macro so makes the parsing somewhat tricky,
    # for now they are statically defined
    macro_functions = {
        "t_set_auto_inv_100": "int state",
        "t_set_disable_6xx": "int state",
        "t_set_disable_failover": "int state",
        "t_set_no_e2e_cancel_reason": "int state",
        "t_set_disable_internal_reply": "int state"
    }

    # These files export the KEMI functions in a special way so we map them manually
    # TODO: Discuss with @miconda if core/HDR/pv/x should be added as well or not
    special_exports = [
        {"filename": "kemi.c", "export": "_sr_kemi_core", "folder": "/src/core/"},
        {"filename": "kemi.c", "export": "_sr_kemi_hdr", "folder": "/src/core/"},
        {"filename": "kemi.c", "export": "_sr_kemi_pv", "folder": "/src/core/"},
        #{"filename": "app_lua_mod.c", "export": "sr_kemi_app_lua_rpc_exports", "folder": "/modules/app_lua/"}
    ]

    def generate_kemi_export_list(self, source_path):
        files = self.list_c_files_in_directory(source_path)
        lists = []

        for file in files:
            with open(file, 'r', encoding='utf-8', errors='ignore') as f:
                lines = f.readlines()

            export_name = self.find_c_file_kemi_export(file, lines)
            if export_name:
                export_functions = self.extract_c_file_kemi_export_lines(file, lines, export_name)
                lists = lists + export_functions
                print ("Found ", len(export_functions), "functions", "Total:", len(lists))

        # Handle some special files separately
        for elem in self.special_exports:
            file = source_path + elem["folder"] + elem["filename"]
            with open(file) as f:
                lines = f.readlines()
            lists = lists + self.extract_c_file_kemi_export_lines(file, lines, elem["export"])

        return lists

    def find_c_file_kemi_export(self, filename, lines):
        param = None

        for line in lines:
            if line.find("sr_kemi_modules_add") >= 0:
                line = line.lstrip(" ")
                line = line.lstrip("\t")
                if line.find("sr_kemi_modules_add") == 0:
                    param = line[line.find("(") + 1:line.find(")")]
                    print ("INFO: ---- Found export", filename, param)
                    break
                else:
                    if line != "int sr_kemi_modules_add(sr_kemi_t *klist)\n":
                        print ("ERR: Possible error at line: ", filename, line)
                        exit()

        return param

    def extract_c_file_kemi_export_lines(self, filename, lines, export_name):
        list_functions = []
        find_start = True

        for line in lines:
            if find_start and line.find("static sr_kemi_t " + export_name + "[]") >= 0:
                find_start = False
            elif not find_start:
                if line.find("};") >= 0:
                    break
                line = line.lstrip(" \t")
                line = line.rstrip()
                list_functions.append(line)

        if len(list_functions) < 1:
            print ("ERR: Couldn't parse file for exported functions: ", export_name)
            exit()

        parsed_list = self.parse_kemi_export_c_lines(filename, list_functions)

        return parsed_list

    def parse_kemi_export_c_lines(self, filename, lines):
        elements = []
        function_lines = []
        temp_function = ""

        # We look for str_init which would be the start of each exported function
        for line in lines:
            if line.find("str_init") >= 0:
                if temp_function != "":
                    function_lines.append(temp_function)
                    temp_function = ""
            temp_function += line

        if temp_function != "":
            function_lines.append(temp_function)

        # Now we parse each exported function to extract its declaration
        for func in function_lines:
            function_lines_split = func.split(",{")

            if len(function_lines_split) < 2:
                print ("ERR: Incorrect function line", func)
                exit()

            declarations = function_lines_split[0].split(",")
            params = function_lines_split[1]

            # Extract the content from the definitions
            val_module = declarations[0]
            val_module = val_module[val_module.find('("') + 2:val_module.find('")')]
            val_function = declarations[1]
            val_function = val_function[val_function.find('("') + 2:val_function.find('")')]

            if declarations[2] == "SR_KEMIP_INT":
                val_return = "int"
            elif declarations[2] == "SR_KEMIP_STR":
                val_return = "string"
            elif declarations[2] == "SR_KEMIP_NONE":
                val_return = "void"
            elif declarations[2] == "SR_KEMIP_BOOL":
                val_return = "bool"
            elif declarations[2] == "SR_KEMIP_XVAL":
                val_return = "xval"
            else:
                print("ERR: Invalid return value", declarations[2], func)
                exit()

            val_c_function = declarations[3].strip()

            # Count how many parameters the KEMI C function expects
            val_params = []
            itr = 0
            for val in params.rstrip("},").split(","):
                itr += 1
                # KEMI function has a maximum of 6 params
                if itr > 6:
                    break
                pm = val.strip()
                if pm == "SR_KEMIP_INT":
                    val_params.append("int")
                elif pm == "SR_KEMIP_STR":
                    val_params.append("str")
                elif pm == "SR_KEMIP_NONE":
                    continue
                else:
                    print("Invalid return value", declarations[2], func)
                    exit()

            if itr != 6:
                print("ERR: Couldn't iterate the params: ", params)
                exit()

            param_string = self.find_c_function_params(filename, val_c_function, val_return)
            param_string = self.prettify_params_list(val_function, param_string, val_params)

            elements.append({"module": val_module, "name": val_function, "ret": val_return, "params": param_string})

        return elements

    def prettify_params_list(self, function_name, function_declaration, kemi_types):
        # Validate the quantity and types of declaration vs export
        if function_declaration == "" and len(kemi_types) == 0:
            return ""

        params = function_declaration.split(",")

        if params[0].find("sip_msg_t") >= 0 or params[0].find("struct sip_msg") >= 0:
            params.pop(0)

        if len(params) != len(kemi_types):
            print("ERR: Mismatching quantity of params. Declaration for", function_name, ":", function_declaration, "KEMI:", kemi_types)
            exit()

        for declared, type in zip(params, kemi_types):
            declared = declared.replace("*", "")
            declared = declared.strip().split(" ")[0]
            if declared != type:
                print("ERR: Mismatching type of params for", function_name, ":", function_declaration, " | ", kemi_types, " | Declared: ", declared, " Type: ", type)
                exit()

        param_string = ""

        for param in params:
            param = param.strip()
            param = param.replace("*", "")
            if param[:3] == "str":
                temp = param.split(" ")
                param = "p_" + temp[1]
            param_string += param + ", "

        # Clean up the presentation of the params
        param_string = param_string.rstrip(", ")
        return param_string

    def find_c_function_params(self, filename, function_name, return_type):
        # First we try with the same file to find the declaration
        param_string = self.search_file_for_function_declaration(filename, function_name, return_type)
        # If we couldn't find it, let's try all files in the same folder as the first file
        if param_string:
            return param_string
        else:
            files = self.list_c_files_in_directory(os.path.dirname(filename))
            for file in files:
                param_string = self.search_file_for_function_declaration(file, function_name, return_type)
                if param_string:
                    return param_string

        if function_name in self.macro_functions:
            return self.macro_functions[function_name]

        print("ERR: Couldn't find the function declaration", filename, function_name, return_type)
        exit()

    def search_file_for_function_declaration(self, filename, function_name, return_type):
        # print "Searching file", filename, "for", function_name
        with open(filename, 'r', encoding='utf-8', errors='ignore') as f:
            lines = f.readlines()

        param_string = None
        found = False
        temp_string = ""
        return_match = return_type

        # KEMI has some magic where most functions actually return INTs but KEMI maps them to void/bool
        if return_type == "void" or return_type == "bool":
            return_match = "int"

        if return_type == "xval":
            return_match = "sr_kemi_xval_t([ \t])*\*"

        # Look for declarations in format:    static? return_type function_name(
        r = re.compile("^(?:static )?" + return_match + "[ \t]*(" + function_name + ")[ \t]*\(")
        for line in lines:
            m = r.match(line)
            if m:
                found = True
            if found:
                temp_string += line
                if line.find("{") >= 0:
                    param_string = temp_string[temp_string.find('(') + 1:temp_string.find(')')]
                    break

        return param_string

    def list_c_files_in_directory(self, path):
        matches = []
        for root, dirnames, filenames in os.walk(path):
            for filename in fnmatch.filter(filenames, '*.c'):
                matches.append(os.path.join(root, filename))
        return matches


if __name__ == "__main__":
    try:
        if not os.path.isdir(sys.argv[1]):
            raise Exception('Not a valid directory')
    except:
        print("Please provide the path to the Kamailio src folder as the first argument")
        exit()

    print("Parsing the source")
    parser = KemiFileExportParser()
    data = parser.generate_kemi_export_list(sys.argv[1].rstrip("/"))
    fgen = ModuleObjGenerator()
    fgen.execute(data)
    print("Done")
